<?php

namespace FiloBlu\Refilo\Model;

use Exception;
use FiloBlu\Newsletter\Model\SubscriberInfoFactory;
use FiloBlu\Refilo\Api\NewsletterManagementInterface;
use FiloBlu\Core\Api\StoreCalculatorResolverInterface;
use Magento\Customer\Api\AccountManagementInterface as CustomerAccountManagement;
use Magento\Customer\Api\CustomerRepositoryInterface;
use Magento\Customer\Model\Customer;
use Magento\Customer\Model\Session;
use Magento\Framework\App\Action\Context;
use Magento\Framework\App\Config\ScopeConfigInterface;
use Magento\Framework\DataObject;
use Magento\Framework\Exception\LocalizedException;
use Magento\Framework\Exception\NoSuchEntityException;
use Magento\Framework\Module\Manager;
use Magento\Framework\ObjectManagerInterface;
use Magento\Framework\Phrase;
use Magento\Framework\Validator\EmailAddress;
use Magento\Newsletter\Model\Subscriber;
use Magento\Newsletter\Model\SubscriberFactory;
use Magento\Store\Api\StoreRepositoryInterface;
use Magento\Store\Model\ScopeInterface;
use Magento\Store\Model\StoreManagerInterface;
use Throwable;
use Zend_Validate;

/**
 * {@inheritDoc}
 */
class NewsletterManagement implements NewsletterManagementInterface
{

    const MAILCHIMP_DOB_FORMAT = 'd/m/Y';
    const MAILCHIMP_BIRTHDAY_FORMAT = 'm/d';

    /**
     * @var ObjectManagerInterface
     */
    protected $_objectManager;

    /**
     * @var StoreManagerInterface
     */
    protected $_storeManager;

    /**
     * @var Session
     */
    protected $_customerSession;

    /**
     * @var CustomerAccountManagement
     */
    protected $customerAccountManagement;

    /**
     * @var SubscriberFactory
     */
    protected $subscriberFactory;
    /**
     * @var SubscriberInfoFactory
     */
    protected $_subscriber_info;

    /**
     * @var Customer
     */
    protected $_customer;

    /**
     * @var Manager
     */
    protected $_moduleManager;
    /**
     * @var Manager
     */
    protected $moduleManager;
    /**
     * @var StoreRepositoryInterface
     */
    private $storeRepository;
    /**
     * @var Context
     */
    private $context;
    /**
     * @var CustomerRepositoryInterface
     */
    private $customerRepositoryInterface;
    /**
     * @var StoreCalculatorResolverInterface
     */
    private $storeCalculatorResolverInterface;

    /**
     * NewsletterManagement constructor.
     * @param Context $context
     * @param StoreManagerInterface $storeManager
     * @param Session $customerSession
     * @param CustomerAccountManagement $customerAccountManagement
     * @param SubscriberFactory $subscriberFactory
     * @param Customer $_customer
     * @param SubscriberInfoFactory $subscriber_info
     * @param Manager $moduleManager
     * @param StoreRepositoryInterface $storeRepository
     * @param CustomerRepositoryInterface $customerRepositoryInterface
     * @param StoreCalculatorResolverInterface $storeCalculatorResolverInterface
     */
    public function __construct(
        Context $context,
        StoreManagerInterface $storeManager,
        Session $customerSession,
        CustomerAccountManagement $customerAccountManagement,
        SubscriberFactory $subscriberFactory,
        Customer $_customer,
        SubscriberInfoFactory $subscriber_info,
        Manager $moduleManager,
        StoreRepositoryInterface $storeRepository,
        CustomerRepositoryInterface $customerRepositoryInterface,
        StoreCalculatorResolverInterface $storeCalculatorResolverInterface
    ) {
        $this->_objectManager = $context->getObjectManager();
        $this->_storeManager = $storeManager;
        $this->_customerSession = $customerSession;
        $this->customerAccountManagement = $customerAccountManagement;
        $this->subscriberFactory = $subscriberFactory;
        $this->_customer = $_customer;
        $this->_subscriber_info = $subscriber_info;
        $this->moduleManager = $moduleManager;
        $this->storeRepository = $storeRepository;
        $this->context = $context;
        $this->customerRepositoryInterface = $customerRepositoryInterface;
        $this->storeCalculatorResolverInterface = $storeCalculatorResolverInterface;
    }

    /**
     * @param string|null $email
     * @param string|null $source
     * @param string|null $firstname
     * @param string|null $lastname
     * @param string|null $dob
     * @param string|null $gender
     * @param string|null $country
     * @param string|null $collections
     * @param string|null $phone
     * @param string|null $language
     * @param string|null $website_name
     * @param string|null $store
     * @param string|null $birthday
     * @param string|null $type
     * @return \Magento\Framework\DataObject
     */
    public function subscribe(
        $email,
        $source,
        $firstname = null,
        $lastname = null,
        $dob = null,
        $gender = null,
        $country = null,
        $collections = null,
        $phone = null,
        $language = null,
        $website_name = null,
        $store = null,
        $birthday = null,
        $type = null
    ) {
        $success = false;
        $subscriber = null;

        // Try to calculate store if missing (but I have the country)
        if (!$store && $country){
            $storeCalculatorFromCountry = $this->storeCalculatorResolverInterface->get('fromCountry');
            $store = ($storeCalculatorFromCountry->calculate($country))->getCode();
        }

        $storeByCode = $this->storeRepository->getActiveStoreByCode($store);
        $existingCustomer = $this->customerExists($email,$storeByCode->getWebsiteId());

        // Calculate type if missing
        if (!$type){
            $type = (($existingCustomer!==false) ? 'CUSTOMER' : 'GUEST');
        }

        // Calculate language if missing
        if (!$language){
            $locale = $this->_objectManager->get(ScopeConfigInterface::class)->getValue(
                'general/locale/code',
                ScopeInterface::SCOPE_STORE,
                $storeByCode->getCode()
            );
            $language = $locale;
        }

        // Calculate website name if missing
        if (!$website_name){
            $website_name = $storeByCode->getWebsite()->getName();
        }

        // Calculate dob and birthday if missing
        if ($dob){
            $date = trim(str_replace('00:00:00','',$dob));
            if ($this->validateDate($date)){
                $date_obj = \DateTime::createFromFormat('Y-m-d', $date);
                $dob = $date_obj->format(self::MAILCHIMP_DOB_FORMAT);
                $birthday = $date_obj->format(self::MAILCHIMP_BIRTHDAY_FORMAT);
            }
        }

        $data = [
            'source'        => $source,
            'firstname'     => $firstname,
            'lastname'      => $lastname,
            'dob'           => $dob,
            'country'       => $country,
            'gender'        => $gender,
            'collections'   => $collections,
            'language'      => $language,
            'website_name'  => $website_name,
            'phone'         => $phone,
            'type'          => $type,
            'birthday'      => $birthday
        ];

        try {
            $this->validateEmailFormat($email);
            $this->validateGuestSubscription();

            $subscriber = $this->subscriberFactory->create()->loadByEmail($email);
            if ($subscriber->getId()
                && $subscriber->getSubscriberStatus() == Subscriber::STATUS_SUBSCRIBED
            ) {
                throw new LocalizedException(
                    __('This email address is already subscribed.')
                );
            }

            foreach ($data as $name => $value){
                $subscriber->setData("filoblu_{$name}",$value);
            }

            $this->_storeManager->setCurrentStore($storeByCode->getId());
//            $subscriber->setStoreId($storeByCode->getId());

            if($existingCustomer!==false) {
                $status = $subscriber->subscribeCustomerById($existingCustomer->getId());
            }
            else {
                $status = $subscriber->subscribe($email);
            }

            if ($status == Subscriber::STATUS_NOT_ACTIVE) {
                $success = true;
                $message = __('The confirmation request has been sent.');
            } else {
                $success = true;
                $message = __('Thank you for your subscription.');
            }
        } catch (LocalizedException $e) {
            $message = __('There was a problem with the subscription: %1', $e->getMessage());
        } catch (Exception $e) {
            $message = __('There was a problem with the subscription: %1', $e->getMessage());
        }

        try {
            if ($email) {
                $data['subscriber_id'] = $subscriber->getId();
                $this->setSubscriberInfo($data);
            }
        } catch (Throwable $e) {
            $message = __('There was a problem with the subscription: %1', $e->getMessage());
        }

        $data = [
            'success' => $success,
            'message' => $message,
        ];

        $result = new DataObject();
        $result->setData($data);
        return $result;
    }

    /**
     * Validates the format of the email address
     * Reference: vendor/magento/module-newsletter/Controller/Subscriber/NewAction.php
     *
     * @param string $email
     * @return void
     * @throws LocalizedException
     */
    protected function validateEmailFormat($email)
    {
        if (!Zend_Validate::is($email, EmailAddress::class)) {
            throw new LocalizedException(__('Please enter a valid email address.'));
        }
    }

    /**
     * Validates that if the current user is a guest, that they can subscribe to a newsletter.
     * Reference: vendor/magento/module-newsletter/Controller/Subscriber/NewAction.php
     *
     * @return void
     * @throws LocalizedException
     */
    protected function validateGuestSubscription()
    {
        if ($this->_objectManager->get(ScopeConfigInterface::class)
                ->getValue(
                    Subscriber::XML_PATH_ALLOW_GUEST_SUBSCRIBE_FLAG,
                    ScopeInterface::SCOPE_STORE
                ) != 1
        ) {
            throw new LocalizedException(
                __(
                    'subscription denied for guests.'
                )
            );
        }
    }

    public function setSubscriberInfo($data)
    {
        if ($this->moduleManager->isEnabled('FiloBlu_Newsletter')) {
            $subscribe = $this->_subscriber_info->create();
            $subscribe->addData($data);
            $subscribe->save();
        }
    }

    /**
     * @param $email
     * @throws LocalizedException
     * @throws NoSuchEntityException
     */
    public function unsubscribe($email)
    {
        $subscriber = $this->subscriberFactory->create()->loadByEmail($email);
        if ($subscriber->getId()) {
            $subscriber->unsubscribe();
        }

        throw new NoSuchEntityException(new Phrase('Subscriber not found'));
    }

    /**
     * Validates that the email address isn't being used by a different account.
     * Reference: vendor/magento/module-newsletter/Controller/Subscriber/NewAction.php
     *
     * @param string $email
     * @return void
     * @throws LocalizedException
     */
    protected function validateEmailAvailable($email)
    {
        $websiteId = $this->_storeManager->getStore()->getWebsiteId();

        if (!$this->customerAccountManagement->isEmailAvailable($email, $websiteId)) {
            throw new LocalizedException(
                __('This email address is already assigned to another user.')
            );
        }
    }

    /**
     * @param $email
     * @param null $websiteId
     * @return false|\Magento\Customer\Api\Data\CustomerInterface
     */
    public function customerExists($email, $websiteId = null)
    {
        try {
            $customer = $this->customerRepositoryInterface->get($email,$websiteId);
        } catch (\Exception $e) {
            return false;
        }
        return $customer;
    }

    /**
     * @param $date
     * @param string $format
     * @return bool
     */
    function validateDate($date, $format = 'Y-m-d')
    {
        $d = \DateTime::createFromFormat($format, $date);
        // The Y ( 4 digits year ) returns TRUE for any integer with any number of digits so changing the comparison from == to === fixes the issue.
        return $d && $d->format($format) === $date;
    }
}
